#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2008-2014 University of Dundee & Open Microscopy Environment.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

''' Helper functions for views that handle object trees '''

import omero
from omero.rtypes import rlong, rlist, rstring, unwrap
from django.conf import settings
from datetime import datetime


def parse_permissions_css(permissions, ownerid, conn):
    ''' Parse numeric permissions into a string of space separated
        CSS classes.

        @param permissions Permissions to parse
        @type permissions L{omero.rtypes.rmap}
        @param ownerid Owner Id for the object having Permissions
        @type ownerId Integer
        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
    '''
    restrictions = ('canEdit', 'canAnnotate', 'canLink', 'canDelete')
    permissionsCss = [r for r in restrictions if permissions.get(r)]
    if ownerid == conn.getUserId():
        permissionsCss.append("canChgrp")
    return ' '.join(permissionsCss)


def _marshal_experimenter(conn, row):
    ''' Given an Experimenter row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * omeName (rstring)
          * firstName (rstring)
          * lastName (rstring)
          * email (rstring)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Experimenter row to marshal
        @type row L{list}
    '''
    experimenter_id, ome_name, first_name, last_name, email = row
    experimenter = dict()
    experimenter['id'] = unwrap(experimenter_id)
    experimenter['omeName'] = unwrap(ome_name)
    experimenter['firstName'] = unwrap(first_name)
    experimenter['lastName'] = unwrap(last_name)
    # Email is not mandatory
    if email:
        experimenter['email'] = unwrap(email)
    return experimenter


def marshal_experimenters(conn, group_id=None, page=0):
    ''' Marshals experimenters, possibly filtered by group.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param group_id The Group ID to filter results by or `None` if we are
        not to filter by a specific group.
        @type group_id L{long}
    '''
    experimenters = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if group_id is not None:
        params.add('gid', rlong(group_id))
        where_clause = 'join experimenter.groupExperimenterMap grexp ' \
                       'where grexp.parent.id = :gid'
    qs = conn.getQueryService()
    q = """
        select experimenter.id,
               experimenter.omeName,
               experimenter.firstName,
               experimenter.lastName,
               experimenter.email
        from Experimenter experimenter %s
        order by lower(experimenter.omeName)
        """ % (where_clause)
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        experimenters.append(_marshal_experimenter(conn, e[0:5]))
    return experimenters

def marshal_experimenter(conn, experimenter_id):
    ''' Marshals experimenter.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter ID to get details for
        @type experimenter_id L{long}
    '''
    params = omero.sys.ParametersI()
    params.add('id', rlong(experimenter_id))
    qs = conn.getQueryService()
    q = """
        select experimenter.id,
               experimenter.omeName,
               experimenter.firstName,
               experimenter.lastName,
               experimenter.email
        from Experimenter experimenter
        where experimenter.id = :id
        """
    rows = qs.projection(q, params, conn.SERVICE_OPTS)
    if len(rows) != 1:
        # TODO Throw exception?
        pass
    return _marshal_experimenter(conn, rows[0][0:5])

def _marshal_project(conn, row):
    ''' Given a Project row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * child_count (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Project row to marshal
        @type row L{list}
    '''
    project_id, name, owner_id, permissions, child_count = row
    project = dict()
    project['id'] = unwrap(project_id)
    project['name'] = unwrap(name)
    project['isOwned'] = unwrap(owner_id) == conn.getUserId()
    project['childCount'] = unwrap(child_count)
    project['permsCss'] = \
        parse_permissions_css(permissions, unwrap(owner_id), conn)
    return project

def marshal_projects(conn, experimenter_id, page=None):
    ''' Marshals projects for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Projects for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
    '''
    projects = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'where project.details.owner.id = :id'
    qs = conn.getQueryService()

    q = """
        select project.id,
               project.name,
               project.details.owner.id,
               project.details.permissions,
               (select count(id) from ProjectDatasetLink pdl
                where pdl.parent = project.id)
        from Project project
        %s
        order by lower(project.name)
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        projects.append(_marshal_project(conn, e[0:5]))
    return projects

def _marshal_dataset(conn, row):
    ''' Given a Dataset row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * child_count (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Dataset row to marshal
        @type row L{list}
    '''
    dataset_id, name, owner_id, permissions, child_count = row
    dataset = dict()
    dataset['id'] = unwrap(dataset_id)
    dataset['name'] = unwrap(name)
    dataset['isOwned'] = unwrap(owner_id) == conn.getUserId()
    dataset['childCount'] = unwrap(child_count)
    dataset['permsCss'] = \
        parse_permissions_css(permissions, unwrap(owner_id), conn)
    return dataset

def marshal_datasets(conn, experimenter_id=None, project_id=None, page=None):
    ''' Marshal datasets for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Datasets for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
    '''
    datasets = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'and dataset.details.owner.id = :id'
    qs = conn.getQueryService()
    q = """
        select dataset.id,
               dataset.name,
               dataset.details.owner.id,
               dataset.details.permissions,
               (select count(id) from DatasetImageLink dil
                 where dil.parent=dataset.id)
               from Dataset dataset
        """

    # If this is a query to get datasets from a parent project
    if project_id:
        params.add('pid', rlong(project_id))
        q+= """
            join dataset.projectLinks plink
            where plink.parent.id = :pid
            """
    # Else this is a query to get datasets with no parent project
    else:
        q+= """
            where not exists (
                select pdlink from ProjectDatasetLink as pdlink
                where pdlink.child = dataset.id
            )
            """

    q+= """
        %s
        order by lower(dataset.name)
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        datasets.append(_marshal_dataset(conn, e[0:5]))
    return datasets

def _marshal_image(conn, row, row_pixels = None):
    ''' Given an Image row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * fileset_id (rlong)

        May also take a row_pixels (list) if X,Y,Z dimensions are loaded
          * pixels.sizeX (rlong)
          * pixels.sizeY (rlong)
          * pixels.sizeZ (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Image row to marshal
        @type row L{list}
        @param row_pixels The Image row pixels data to marshal
        @type row_pixels L{list}
    '''
    image_id, name, owner_id, permissions, fileset_id = row
    image = dict()
    image['id'] = unwrap(image_id)
    image['name'] = unwrap(name)
    image['isOwned'] = unwrap(owner_id) == conn.getUserId()
    image['permsCss'] = parse_permissions_css(permissions, unwrap(owner_id), conn)
    image['filesetId'] = unwrap(fileset_id)
    if row_pixels:
        sizeX, sizeY, sizeZ = row_pixels
        image['sizeX'] = unwrap(sizeX)
        image['sizeY'] = unwrap(sizeY)
        image['sizeZ'] = unwrap(sizeZ)

    return image

def marshal_images(conn, experimenter_id=None, dataset_id=None, share_id=None,
                   page=None, load_pixels=False):
    ''' Marshal images for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Images for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
        @param dataset_id The Dataset ID to marhsal Images for or `None`
        if we are not to filter by a specific dataset
        @type dataset_id L{long}
        @param share_id The Share ID to marhsal Images for or `None`
        if we are not getting images for a share
        @type share_id L{long}
        @param load_pixels Whether to load the X,Y,Z dimensions
        @type load_pixels Boolean
    '''
    images = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'and image.details.owner.id = :id'
    qs = conn.getQueryService()

    q = """
        select image.id,
               image.name,
               image.details.owner.id,
               image.details.permissions,
               image.fileset.id
        """

    if load_pixels:
        q += """
             ,
             pix.sizeX,
             pix.sizeY,
             pix.sizeZ
             """

    q += 'from Image image '

    if load_pixels:
        q += 'join image.pixels pix '

    # If this is a query to get images from a parent dataset
    if dataset_id is not None:
        params.add('did', rlong(dataset_id))
        q+= 'join image.datasetLinks dlink ' \
            'where dlink.parent.id = :did '

    # If this is a query to get images in a share
    elif share_id is not None:
        # Get the contents of the blob which contains the images in the share
        image_rids = [x._obj.getId() for x in conn.getContents(share_id)
                      if isinstance(x._obj, omero.model.ImageI)]

        # If there are no images in the share, don't bother querying
        if not image_rids:
            return images

        params.add('iids', rlist(image_rids))
        q+= """
            where image.id in (:iids)
            """

    # Else this is a query to get images with no parent datasets (orphans)
    # TODO An orphan should be considered to be when the user has no
    # DatasetImageLinks not when there are no DatasetImageLinks at all. At
    # the moment if another user has a DatasetImageLink using your image
    # then it does not appear under any dataset in your list or in your
    # orphans
    else:
        q+= """
            where not exists (
                select dilink from DatasetImageLink as dilink
                where dilink.child = image.id
            """
        if experimenter_id is not None:
            q+= '''
                and dilink.parent.details.owner.id = :id
                '''

        q+= ') '

        # Also discount any images which are part of a screen. No need to
        # take owner into account on this because we don't want them in
        # orphans either way
        q+= """
            and not exists (
                select ws from WellSample ws
                where ws.image.id = image.id
            )
            """

    q+= """
        %s
        order by lower(image.name)
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        kwargs = {'conn':conn, 'row':e[0:5]}
        if load_pixels:
            kwargs['row_pixels'] = e[5:8]

        images.append(_marshal_image(**kwargs))
    return images

def _marshal_screen(conn, row):
    ''' Given a Screen row (list) marshals it into a dictionary.  Order and
        type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * child_count (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Screen row to marshal
        @type row L{list}
    '''

    screen_id, name, owner_id, permissions, child_count = row
    screen = dict()
    screen['id'] = unwrap(screen_id)
    screen['name'] = unwrap(name)
    screen['isOwned'] = unwrap(owner_id) == conn.getUserId()
    screen['childCount'] = unwrap(child_count)
    screen['permsCss'] = \
        parse_permissions_css(permissions, unwrap(owner_id), conn)
    return screen

def marshal_screens(conn, experimenter_id=None, page=None):
    ''' Marshals screens for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Screens for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
    '''
    screens = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'where screen.details.owner.id = :id'
    qs = conn.getQueryService()
    q = """
        select screen.id,
               screen.name,
               screen.details.owner.id,
               screen.details.permissions,
               (select count(spl.id) from ScreenPlateLink spl
                where spl.parent=screen.id)
               from Screen screen
               %s
               order by lower(screen.name)
        """ % where_clause

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        screens.append(_marshal_screen(conn, e[0:5]))

    return screens

def _marshal_plate(conn, row):
    ''' Given a Plate row (list) marshals it into a dictionary.  Order and
        type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * child_count (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Plate row to marshal
        @type row L{list}
    '''

    plate_id, name, owner_id, permissions, child_count = row
    plate = dict()
    plate['id'] = unwrap(plate_id)
    plate['name'] = unwrap(name)
    plate['isOwned'] = unwrap(owner_id) == conn.getUserId()
    plate['childCount'] = unwrap(child_count)
    plate['permsCss'] = \
        parse_permissions_css(permissions, unwrap(owner_id), conn)
    return plate

def marshal_plates(conn, experimenter_id=None, screen_id=None, page=None):
    ''' Marshals plates for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Plates for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
        @param screen_id The Screen ID to marhsal Plates for or `None`
        if we are not to filter by a specific screen
        @type screen_id L{long}
    '''
    plates = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'and plate.details.owner.id = :id'
    qs = conn.getQueryService()
    q = """
        select plate.id,
               plate.name,
               plate.details.owner.id,
               plate.details.permissions,
               (select count(pa.id) from PlateAcquisition pa
                where pa.plate.id=plate.id)
        from Plate plate
        """

    # If this is a query to get plates from a parent screen
    if screen_id is not None:
        params.add('sid', rlong(screen_id))
        q+= """
            join plate.screenLinks slink
            where slink.parent.id = :sid
            """
    # Else this is a query to get datasets with no parent project
    else:
        q+= """
            where not exists (
                select splink from ScreenPlateLink as splink
                where splink.child = plate.id
            )
            """

    q+= """
        %s
        order by lower(plate.name)
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        plates.append(_marshal_plate(conn, e[0:5]))

    return plates

def _marshal_plate_acquisition(conn, row):
    ''' Given a PlateAcquisition row (list) marshals it into a dictionary.
        Order and type of columns in row is:
          * id (rlong)
          * name (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * startTime (rtime)
          * endTime (rtime)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The PlateAcquisition row to marshal
        @type row L{list}
    '''

    pa_id, name, owner_id, permissions, start_time, end_time = row
    plate_acquisition = dict()
    plate_acquisition['id'] = unwrap(pa_id)

    # If there is no defined name, base it on the start/end time if that
    # exists or finally default to an id based name
    if name is not None:
        plate_acquisition['name'] = unwrap(name)
    elif start_time is not None and end_time is not None:
        start_time = datetime.utcfromtimestamp(unwrap(start_time) / 1000.0)
        end_time = datetime.utcfromtimestamp(unwrap(end_time) / 1000.0)
        plate_acquisition['name'] = '%s - %s' % (start_time, end_time)
    else:
        plate_acquisition['name'] = 'Run %d' % unwrap(pa_id)

    plate_acquisition['isOwned'] = unwrap(owner_id) == conn.getUserId()
    plate_acquisition['permsCss'] = \
        parse_permissions_css(permissions, unwrap(owner_id), conn)
    return plate_acquisition

def marshal_plate_acquisitions(conn, plate_id, page=None):
    # TODO Think about this for this/other cases:
    # It is default behaviour to return orphans if there is no object_id
    # specified. In this case, 'plate_id'. In the case of PlateAcquisitions
    # there can be no orphans so these are gone from the interface. Also,
    # the owner of the PlateAcquisition will be the same as for the plate_id
    # so it does not make sense to have experimenter_id
    ''' Marshals PlateAcquisitions ('runs') for a given plate.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param plate_id The Plate ID to marshal PlateAcquisitions
        @type plate_id L{long}
    '''
    plate_acquisitions = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    params.add('pid', rlong(plate_id))
    qs = conn.getQueryService()
    q = """
        select pa.id,
               pa.name,
               pa.details.owner.id,
               pa.details.permissions,
               pa.startTime,
               pa.endTime
        from PlateAcquisition pa
        where pa.plate.id = :pid
        """

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        plate_acquisitions.append(_marshal_plate_acquisition(conn, e[0:6]))

    return plate_acquisitions

def marshal_orphaned(conn, experimenter_id, page=None):
    ''' Marshals orphaned container for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        orphaned container for.
        @type experimenter_id L{long}
    '''
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    params.addId(experimenter_id)
    qs = conn.getQueryService()

    # Count all the images that do not have Datasets as parents or are
    # not in a WellSample
    q = '''
        select count(image.id) from Image image
        where image.details.owner.id = :id
        and not exists (
            select dilink from DatasetImageLink as dilink
            where dilink.child.id = image.id
            and dilink.parent.details.owner.id = :id
        )
        and not exists (
                select ws from WellSample ws
                where ws.image.id = image.id
        )
        '''

    count = unwrap(qs.projection(q, params, conn.SERVICE_OPTS)[0][0])
    orphaned = dict()
    # In orphans, record the id as the experimenter
    orphaned['id'] = experimenter_id
    orphaned['childCount'] = count
    return orphaned

def _marshal_tag(conn, row):
    ''' Given a Tag row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * text_value (rstring)
          * description (rstring)
          * details.owner.id (rlong)
          * details.permissions (dict)
          * namespace (rstring)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Tag row to marshal
        @type row L{list}

    '''
    tag_id, text_value, description, owner_id, permissions, namespace, \
        child_count = row

    tag = dict()
    tag['id'] = unwrap(tag_id)
    tag['value'] = unwrap(text_value)
    tag['description'] = unwrap(description)
    tag['isOwned'] = unwrap(owner_id) == conn.getUserId()
    tag['permsCss'] = parse_permissions_css(permissions, unwrap(owner_id), conn)

    if namespace:
        tag['set'] = True
    else :
        tag['set'] = False

    tag['childCount'] = unwrap(child_count)

    return tag

def marshal_tags(conn, experimenter_id=None, tag_id=None, page=None):
    ''' Marshals tags for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        tags for.
        @type experimenter_id L{long}
        @param tag_id The tag ID to marshal tags for.
        @type tag_id L{long}
    '''
    tags = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    qs = conn.getQueryService()

    # Restricted by the specified tag set
    if tag_id is not None:
        params.add('tid', rlong(tag_id))

        q = '''
            select aalink.child.id,
                   aalink.child.textValue,
                   aalink.child.description,
                   aalink.child.details.owner.id,
                   aalink.child.details.permissions,
                   aalink.child.ns,
                   (select count(aalink2)
                    from AnnotationAnnotationLink aalink2
                    where aalink2.child.class=TagAnnotation
                    and aalink2.parent.id=aalink.child.id)
            from AnnotationAnnotationLink aalink
            where aalink.parent.class=TagAnnotation
            and aalink.child.class=TagAnnotation
            and aalink.parent.id=:tid
            '''

        # Restricted by the specified user
        if experimenter_id is not None:
            params.addId(experimenter_id)
            q+= '''
                and aalink.child.details.owner.id = :id
                '''
    # All
    else:
        q = '''
            select tag.id,
                   tag.textValue,
                   tag.description,
                   tag.details.owner.id,
                   tag.details.permissions,
                   tag.ns,
                   (select count(aalink2)
                    from AnnotationAnnotationLink aalink2
                    where aalink2.child.class=TagAnnotation
                    and aalink2.parent.id=tag.id)
            from TagAnnotation tag
            '''

        # Restricted by the specified user
        if experimenter_id is not None:
            params.addId(experimenter_id)
            q+= '''
                where tag.details.owner.id = :id
                '''

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        tags.append(_marshal_tag(conn, e[0:7]))

    return tags

# TODO This could be built into the standard container marshalling as a filter
# as basically this is just the same as running several of those queries. Park
# this for now, but revisit later
# This also has a slightly different interface to the other marshals in that it
# returns a dictionary of the tagged types. This would also disappear if the
# above marshalling functions had filter functions added as one of those would
# be called each per object type instead of this one for all
def marshal_tagged(conn, experimenter_id, tag_id, page=None):
    ''' Marshals tagged data for a given tag and user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        tags for.
        @type experimenter_id L{long}
        @param tag_id The tag ID to marshal tags for.
        @type tag_id L{long}
    '''
    tagged = {}
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    qs = conn.getQueryService()

    common_clause = '''
                    join obj.annotationLinks alink
                    where alink.child.id=:tid
                    '''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        common_clause+= '''
                        and obj.details.owner.id = :id
                        '''
    common_clause+= '''
                    order by lower(obj.name)
                    '''

    params.add('tid', rlong(tag_id))

    # Projects
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               (select count(id) from ProjectDatasetLink pdl
                where pdl.parent = obj.id)
        from Project obj
        %s
        ''' % common_clause

    projects = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        projects.append(_marshal_project(conn, e[0:5]))
    tagged['projects'] = projects

    # Datasets
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               (select count(id) from DatasetImageLink dil
                 where dil.parent=obj.id)
        from Dataset obj
        %s
        ''' % common_clause

    datasets = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        datasets.append(_marshal_dataset(conn, e[0:5]))
    tagged['datasets'] = datasets

    # Images
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               obj.fileset.id
        from Image obj
        %s
        ''' % common_clause

    images = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        images.append(_marshal_image(conn, e[0:6]))
    tagged['images'] = images

    # Screens
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               (select count(spl.id) from ScreenPlateLink spl
                where spl.parent=obj.id)
        from Screen obj
        %s
        ''' % common_clause

    screens = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        screens.append(_marshal_screen(conn, e[0:5]))
    tagged['screens'] = screens

    # Plate
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               (select count(pa.id) from PlateAcquisition pa
                where pa.plate.id=obj.id)
        from Plate obj
        %s
        ''' % common_clause

    plates = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        plates.append(_marshal_plate(conn, e[0:5]))
    tagged['plates'] = plates


    # Plate Acquisitions
    q = '''
        select obj.id,
               obj.name,
               obj.details.owner.id,
               obj.details.permissions,
               obj.startTime,
               obj.endTime
        from PlateAcquisition obj
        %s
        ''' % common_clause

    plate_acquisitions = []
    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        plate_acquisitions.append(_marshal_plate_acquisition(conn, e[0:6]))
    tagged['acquisitions'] = plate_acquisitions

    return tagged

def _marshal_share(conn, row):
    ''' Given a Share row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * details.owner.id (rlong)
          * child_count (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Share row to marshal
        @type row L{list}
    '''
    share_id, owner_id, child_count = row
    share = dict()
    share['id'] = unwrap(share_id)
    share['isOwned'] = unwrap(owner_id) == conn.getUserId()
    share['childCount'] = unwrap(child_count)

    return share

def marshal_shares(conn, experimenter_id=None, page=None):
    ''' Marshal shares for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Shares for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
    '''
    shares = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'and share.owner.id = :id'
    qs = conn.getQueryService()
    q = """
        select share.id,
               share.owner.id,
               share.itemCount
        from Share share
        where share.itemCount > 0
        %s
        order by share.id
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        shares.append(_marshal_share(conn, e[0:3]))
    return shares

def _marshal_discussion(conn, row):
    ''' Given a Discussion row (list) marshals it into a dictionary.  Order
        and type of columns in row is:
          * id (rlong)
          * details.owner.id (rlong)

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param row The Discussion row to marshal
        @type row L{list}
    '''
    discussion_id, owner_id = row
    discussion = dict()
    discussion['id'] = unwrap(discussion_id)
    discussion['isOwned'] = unwrap(owner_id) == conn.getUserId()

    return discussion

def marshal_discussions(conn, experimenter_id=None, page=None):
    ''' Marshal discussion for a given user.

        @param conn OMERO gateway.
        @type conn L{omero.gateway.BlitzGateway}
        @param experimenter_id The Experimenter (user) ID to marshal
        Discussions for or `None` if we are not to filter by a specific user.
        @type experimenter_id L{long}
    '''
    discussions = []
    params = omero.sys.ParametersI()

    # Paging
    if page is not None:
		params.page((page-1) * settings.PAGE, settings.PAGE)

    where_clause = ''
    if experimenter_id is not None:
        params.addId(experimenter_id)
        where_clause = 'and share.owner.id = :id'
    qs = conn.getQueryService()
    q = """
        select share.id,
               share.owner.id
        from Share share
        where share.itemCount = 0
        %s
        order by share.id
        """ % (where_clause)

    for e in qs.projection(q, params, conn.SERVICE_OPTS):
        discussions.append(_marshal_discussion(conn, e[0:2]))
    return discussions

# Really tags and tagsets are mostly the same, they can be dealt with
# seperately like this, but for now deal with as one with an indicator
# def _marshal_tagset(conn, row):
#     ''' Given a TagSet row (list) marshals it into a dictionary.  Order
#         and type of columns in row is:
#           * id (rlong)
#           * text_value (rstring)
#           * description (rstring)
#           * details.owner.id (rlong)
#           * details.permissions (dict)
#           * fileset_id (rlong)

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param row The Tag row to marshal
#         @type row L{list}

#     '''
#     tagset_id, text_value, description, owner_id, permissions = row
#     tagset = dict()
#     tagset['id'] = unwrap(tagset_id)
#     tagset['value'] = unwrap(text_value)
#     tagset['description'] = unwrap(description)
#     tagset['isOwned'] = unwrap(owner_id) == conn.getUserId()
#     tagset['permsCss'] = parse_permissions_css(permissions, unwrap(owner_id), conn)

#     return tagset

# def marshal_tagsets(conn, experimenter_id=None):
#     ''' Marshals tagsets for a given user.

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param experimenter_id The Experimenter (user) ID to marshal
#         tags for.
#         @type experimenter_id L{long}
#     '''
#     tagsets = []
#     params = omero.sys.ParametersI()
#     qs = conn.getQueryService()

#     # Get all tags
#     q = '''
#         select tagset.id,
#                tagset.textValue,
#                tagset.description,
#                tagset.details.owner.id,
#                tagset.details.permissions
#         from TagAnnotation tagset
#         where tagset.ns is not Null
#         '''

#     # Restricted by the specified user
#     if experimenter_id is not None:
#         params.addId(experimenter_id)
#         q+= '''
#             and tagset.details.owner.id = :id
#             '''

#     for e in qs.projection(q, params, conn.SERVICE_OPTS):
#         tagsets.append(_marshal_tag(conn, e[0:5]))

#     return tagsets

# def _marshal_tag(conn, row):
#     ''' Given a Tag row (list) marshals it into a dictionary.  Order
#         and type of columns in row is:
#           * id (rlong)
#           * text_value (rstring)
#           * description (rstring)
#           * details.owner.id (rlong)
#           * details.permissions (dict)
#           * fileset_id (rlong)

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param row The Tag row to marshal
#         @type row L{list}

#     '''
#     tag_id, text_value, description, owner_id, permissions = row
#     tag = dict()
#     tag['id'] = unwrap(tag_id)
#     tag['value'] = unwrap(text_value)
#     tag['description'] = unwrap(description)
#     tag['isOwned'] = unwrap(owner_id) == conn.getUserId()
#     tag['permsCss'] = parse_permissions_css(permissions, unwrap(owner_id), conn)

#     return tag

# def marshal_tags(conn, experimenter_id=None, tagset_id=None):
#     ''' Marshals tags for a given user.

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param experimenter_id The Experimenter (user) ID to marshal
#         tags for.
#         @type experimenter_id L{long}
#         @param tagset_id The tagset ID to marshal tags for.
#         @type tagset_id L{long}
#     '''
#     tags = []
#     params = omero.sys.ParametersI()
#     qs = conn.getQueryService()

#     #select tag, aalink.child
#     #from TagAnnotation tag
#     #join tag.annotationLinks aalink
#     #where aalink.parent.id=53

#     # Get all tags


#     # Restricted by the specified tag set
#     if tagset_id is not None:
#         params.add('tsid', rlong(tagset_id))

#         q = '''
#             select aalink.child.id,
#                    aalink.child.textValue,
#                    aalink.child.description,
#                    aalink.child.details.owner.id,
#                    aalink.child.details.permissions
#             from AnnotationAnnotationLink aalink
#             where aalink.parent.class=TagAnnotation
#             and aalink.child.class=TagAnnotation
#             and aalink.parent.id=53
#             and aalink.child.ns is Null
#             '''

#         # Restricted by the specified user
#         if experimenter_id is not None:
#             params.addId(experimenter_id)
#             q+= '''
#                 and aalink.child.details.owner.id = :id
#                 '''
#     else:
#         q = '''
#             select tag.id,
#                    tag.textValue,
#                    tag.description,
#                    tag.details.owner.id,
#                    tag.details.permissions
#             from TagAnnotation tag
#             where tag.ns is Null
#             '''

#         # Restricted by the specified user
#         if experimenter_id is not None:
#             params.addId(experimenter_id)
#             q+= '''
#                 and tag.details.owner.id = :id
#                 '''

#     for e in qs.projection(q, params, conn.SERVICE_OPTS):
#         tags.append(_marshal_tag(conn, e[0:5]))

#     return tags

# Understanding Screens
# A screen is just a container
# A plate is the exciting bit, it has X*Y wells
# A PlateAcquisition records a plate, a 'run'. A plate may have multiple runs.
# A well is a bucket which may or may not have something in it
# A WellSample puts an image in the bucket
# WellSamples are linked to a Well by some sort of HQL ordered collection so
# the index of each WellSample is the order in which they appear in fields. This
# is not necessarily complete as the WellSamples may have indices of 0,1,2,5,8
# indiciating that these WellSamples are in fields 0,1,2,5 and 8, but not in
# 3,4,6,7 and potentially any above 8
# select p.id, w.id, ws.id, index(ws) from Plate p join p.wells w join w.wellSamples ws
#
# In summary,
# Plate and Well describe the structure
# PlateAcquisition and WellSample record the actual data

# def marshal_screens(conn, experimenter_id=None):
#     ''' Marshals screens and contained plates and aquisitions for a given user.

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param experimenter_id The Experimenter (user) ID to marshal
#         Screens for or `None` if we are not to filter by a specific user.
#         @type experimenter_id L{long}
#     '''
#     screens = list()
#     query_service = conn.getQueryService()
#     params = omero.sys.ParametersI()
#     where_clause = ''
#     if experimenter_id is not None:
#         params.addId(experimenter_id)
#         where_clause = 'where screen.details.owner.id = :id'
#     q = """
#         select screen.id,
#                screen.name,
#                screen.details.owner.id,
#                screen.details.permissions,
#                plate.id,
#                plate.name,
#                plate.details.owner.id,
#                plate.details.permissions,
#                pa.id,
#                pa.name,
#                pa.details.owner.id,
#                pa.details.permissions,
#                pa.startTime,
#                pa.endTime
#                from Screen screen
#                left join screen.plateLinks splink
#                left join splink.child plate
#                left join plate.plateAcquisitions pa
#         %s
#         order by lower(screen.name), lower(plate.name), pa.id
#         """ % (where_clause)
#     for row in query_service.projection(q, params, conn.SERVICE_OPTS):
#         screen_id, screen_name, screen_owner_id, screen_permissions, \
#             plate_id, plate_name, plate_owner_id, plate_permissions, \
#             acquisition_id, acquisition_name, acquisition_owner_id, \
#             acquisition_permissions, acquisition_start_time, \
#             acquisition_end_time = row
#         if len(screens) == 0 or unwrap(screen_id) != screens[-1]['id']:
#             is_owned = unwrap(screen_owner_id) == conn.getUserId()
#             perms_css = parse_permissions_css(
#                 screen_permissions, unwrap(screen_owner_id), conn
#             )
#             screen = {
#                 'id': unwrap(screen_id), 'name': unwrap(screen_name),
#                 'isOwned': is_owned, 'permsCss': perms_css,
#                 'plates': list()
#             }
#             screens.append(screen)
#         screen = screens[-1]
#         if plate_id is not None and (
#                 len(screen['plates']) == 0 or
#                 screen['plates'][-1]['id'] != unwrap(plate_id)
#                 ):
#             screen['plates'].append(_marshal_plate(conn, (
#                 plate_id, plate_name, plate_owner_id, plate_permissions
#             )))
#             screen['plates'][-1]['plateAcquisitionCount'] = 0
#         if acquisition_id is not None:
#             plate = screen['plates'][-1]
#             plate['plateAcquisitions'].append(_marshal_plate_acquisition(conn, (
#                 acquisition_id, acquisition_name, acquisition_owner_id,
#                 acquisition_permissions, acquisition_start_time,
#                 acquisition_end_time
#             )))
#             plate['plateAcquisitionCount'] = len(plate['plateAcquisitions'])
#         screen['childCount'] = len(screen['plates'])
#     return screens


# def marshal_plates(conn, experimenter_id):
#     ''' Marshal plates for a given user.

#         @param conn OMERO gateway.
#         @type conn L{omero.gateway.BlitzGateway}
#         @param experimenter_id The Experimenter (user) ID to marshal
#         Plates for or `None` if we are not to filter by a specific user.
#         @type experimenter_id L{long}
#     '''
#     plates = list()
#     params = omero.sys.ParametersI()
#     where_clause = ''
#     if experimenter_id is not None:
#         params.addId(experimenter_id)
#         where_clause = 'and plate.details.owner.id = :id'
#     query_service = conn.getQueryService()
#     q = """
#         select plate.id,
#                plate.name,
#                plate.details.owner.id,
#                plate.details.permissions,
#                pa.id,
#                pa.name,
#                pa.details.owner.id,
#                pa.details.permissions,
#                pa.startTime,
#                pa.endTime
#                from Plate plate
#                left join plate.plateAcquisitions pa
#         where not exists (
#             select splink from ScreenPlateLink as splink
#             where splink.child = plate.id
#         ) %s
#         order by lower(plate.name), pa.id
#         """ % (where_clause)
#     for row in query_service.projection(q, params, conn.SERVICE_OPTS):
#         plate_id, plate_name, plate_owner_id, plate_permissions, \
#             acquisition_id, acquisition_name, acquisition_owner_id, \
#             acquisition_permissions, acquisition_start_time, \
#             acquisition_end_time = row
#         if len(plates) == 0 or unwrap(plate_id) != plates[-1]['id']:
#             plates.append(_marshal_plate(conn, (
#                 plate_id, plate_name, plate_owner_id, plate_permissions
#             )))
#             plates[-1]['plateAcquisitionCount'] = 0
#         plate = plates[-1]
#         if acquisition_id is not None:
#             plate['plateAcquisitions'].append(_marshal_plate_acquisition(conn, (
#                 acquisition_id, acquisition_name, acquisition_owner_id,
#                 acquisition_permissions, acquisition_start_time,
#                 acquisition_end_time
#             )))
#             plate['plateAcquisitionCount'] = len(plate['plateAcquisitions'])
#     return plates
